/*
 *  $Id: global_ops.c,v 1.8 1994/02/01 21:19:59 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: global_ops.c,v 1.8 1994/02/01 21:19:59 gropp Exp $";
#endif /* lint */

/*D

Global Compute Operations -

D*/

#include <mpi.h>
#include "coll.h"

typedef struct { 
  float re;
  float im; 
} s_complex;

typedef struct { 
  double re;
  double im; 
} d_complex;


void MPI_MAX( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MAX(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MAX(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MAX(a[i],b[i]);
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MAX(a[i],b[i]);
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MAX(a[i],b[i]);
    break;
  }
  default:
    break;
  }
}


void MPI_MAXLOC( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;
  
  /* Half the length, since we are finding the max of only */
  /* the top half.  The bottom half has the location.  */
  len = len / 2;

  /* Perform the operation */
  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] < b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] < b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] < b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] < b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] < b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  default:
    break;
  }
}

void MPIR_Loc_init(sendbuf, count, datatype, rank)
void         *sendbuf;
int           count;
MPI_Datatype  datatype;
int           rank;
{
  int i;

  /* Half the count */
  count = count / 2;

  switch (datatype->dte_type) {
  case MPIR_INT: {
    int *a = (int *)(sendbuf) + count;
    for ( i=0; i<count; i++ )
	  a[i] = rank;
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)(sendbuf) + count;
    for ( i=0; i<count; i++ )
	  a[i] = rank;
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)(sendbuf) + count;
    for ( i=0; i<count; i++ )
	  a[i] = rank;
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)(sendbuf) + count;
    for ( i=0; i<count; i++ )
	  a[i] = rank;
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)(sendbuf) + count;
    for ( i=0; i<count; i++ )
	  a[i] = rank;
    break;
  }
  default:
    break;
  }
}

void MPI_MIN ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MIN(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MIN(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MIN(a[i],b[i]);
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MIN(a[i],b[i]);
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_MIN(a[i],b[i]);
    break;
  }
  default:
    break;
  }
}

void MPI_MINLOC ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  /* Half the length, since we are finding the min of only */
  /* the top half */
  len = len / 2;

  /* Perform the operation */
  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] > b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] > b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] > b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] > b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
	  if (a[i] > b[i]) {
		a[i] = b[i];
		a[i+len] = b[i+len];
	  }
    break;
  }
  default:
    break;
  }
}


#ifndef MPIR_SUM
#define MPIR_SUM(a,b) ((a)+(b))
#endif
void MPI_SUM ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_SUM(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_SUM(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_SUM(a[i],b[i]);
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_SUM(a[i],b[i]);
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_SUM(a[i],b[i]);
    break;
  }
  case MPIR_COMPLEX: {
    s_complex *a = (s_complex *)inoutvec; s_complex *b = (s_complex *)invec;
    for ( i=0; i<len; i++ ) {
      a[i].re = MPIR_SUM(a[i].re ,b[i].re);
      a[i].im = MPIR_SUM(a[i].im ,b[i].im);
    }
    break;
  }
  case MPIR_DOUBLE_COMPLEX: {
    d_complex *a = (d_complex *)inoutvec; d_complex *b = (d_complex *)invec;
    for ( i=0; i<len; i++ ) {
      a[i].re = MPIR_SUM(a[i].re ,b[i].re);
      a[i].im = MPIR_SUM(a[i].im ,b[i].im);
    }
    break;
  }
  default:
    break;
  }
}



#define MPIR_PROD(a,b) ((a)*(b))
void MPI_PROD ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_PROD(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_PROD(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_PROD(a[i],b[i]);
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_PROD(a[i],b[i]);
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_PROD(a[i],b[i]);
    break;
  }
  case MPIR_COMPLEX: {
    s_complex *a = (s_complex *)inoutvec; s_complex *b = (s_complex *)invec;
    for ( i=0; i<len; i++ ) {
	  s_complex c;
	  c.re = a[i].re; c.im = a[i].im;
      a[i].re = c.re*b[i].re - c.im*b[i].im;
      a[i].im = c.im*b[i].re + c.re*b[i].im;
    }
    break;
  }
  case MPIR_DOUBLE_COMPLEX: {
    d_complex *a = (d_complex *)inoutvec; d_complex *b = (d_complex *)invec;
    for ( i=0; i<len; i++ ) {
      d_complex c;
	  c.re = a[i].re; c.im = a[i].im;
      a[i].re = c.re*b[i].re - c.im*b[i].im;
      a[i].im = c.im*b[i].re + c.re*b[i].im;
    }
    break;
  }
  default:
    break;
  }
}



#ifndef MPIR_LAND
#define MPIR_LAND(a,b) ((a)&&(b))
#endif
void MPI_LAND ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LAND(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LAND(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LAND(a[i],b[i]);
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LAND(a[i],b[i]);
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LAND(a[i],b[i]);
    break;
  }
  default:
    break;
  }
}



#ifndef MPIR_BAND
#define MPIR_BAND(a,b) ((a)&(b))
#endif
void MPI_BAND ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BAND(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BAND(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BAND(a[i],b[i]);
    break;
  }
  default:
    break;
  }
}



#ifndef MPIR_LOR
#define MPIR_LOR(a,b) ((a)||(b))
#endif
void MPI_LOR ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LOR(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LOR(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LOR(a[i],b[i]);
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LOR(a[i],b[i]);
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LOR(a[i],b[i]);
    break;
  }
  default:
    break;
  }
}


#ifndef MPIR_BOR
#define MPIR_BOR(a,b) ((a)|(b))
#endif
void MPI_BOR ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BOR(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BOR(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BOR(a[i],b[i]);
    break;
  }
  default:
    break;
  }
}



#ifndef MPIR_LXOR
#define MPIR_LXOR(a,b) (((a)&&(!b))||((!a)&&(b)))
#endif
void MPI_LXOR ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LXOR(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LXOR(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LXOR(a[i],b[i]);
    break;
  }
  case MPIR_FLOAT: {
    float *a = (float *)inoutvec; float *b = (float *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LXOR(a[i],b[i]);
    break;
  }
  case MPIR_DOUBLE: {
    double *a = (double *)inoutvec; double *b = (double *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_LXOR(a[i],b[i]);
    break;
  }
  default:
    break;
  }
}



#ifndef MPIR_BXOR
#define MPIR_BXOR(a,b) ((a)^(b))
#endif
void MPI_BXOR ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
  int i;

  switch (type->dte_type) {
  case MPIR_INT: {
    int *a = (int *)inoutvec; int *b = (int *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BXOR(a[i],b[i]);
    break;
  }
  case MPIR_LONG: {
    long *a = (long *)inoutvec; long *b = (long *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BXOR(a[i],b[i]);
    break;
  }
  case MPIR_SHORT: {
    short *a = (short *)inoutvec; short *b = (short *)invec;
    for ( i=0; i<len; i++ )
      a[i] = MPIR_BXOR(a[i],b[i]);
    break;
  }
  default:
    break;
  }
}


void MPIR_Init_coll_op ( sendbuf, count, datatype, op, rank )
void         *sendbuf;
int           count;
MPI_Datatype  datatype;
void        (*op)();
int           rank;
{
  if ( (op == MPI_MAXLOC) || (op == MPI_MINLOC) )
	MPIR_Loc_init(sendbuf, count, datatype, rank);
  return;
}


/* 
    Here begins the Fortran versions of the combination routines
 */
#if !defined(MPI_cray)
/* Crays use uppercase, no trailing _ for Fortran.  Fortunately, we don't
   need separate routines in that case */
#if defined(MPI_rs6000)
#define mpi_sum_ mpi_sum
#define mpi_max_ mpi_max
#define mpi_min_ mpi_min
#define mpi_prod_ mpi_prod
#define mpi_land_ mpi_land
#define mpi_band_ mpi_band
#define mpi_lor_ mpi_lor
#define mpi_bor_ mpi_bor
#define mpi_lxor_ mpi_lxor
#define mpi_bxor_ mpi_bxor
#endif

void mpi_max_( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_MAX( invec, inoutvec, len, type );
}


void mpi_maxloc_( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_MAXLOC( invec, inoutvec, len, type );
}

void mpi_min_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_MIN( invec, inoutvec, len, type );
}

void mpi_minloc_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_MINLOC ( invec, inoutvec, len, type );
}


void mpi_sum_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_SUM( invec, inoutvec, len, type );
}


void mpi_prod_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_PROD ( invec, inoutvec, len, type );
}



void mpi_land_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_LAND ( invec, inoutvec, len, type );
}



void mpi_band_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_BAND ( invec, inoutvec, len, type );
}



void mpi_lor_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_LOR ( invec, inoutvec, len, type );
}


void mpi_bor_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_BOR ( invec, inoutvec, len, type );
}



void mpi_lxor_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_LXOR ( invec, inoutvec, len, type );
}



void mpi_bxor_ ( invec, inoutvec, len, type )
void *invec, *inoutvec;
int  len;
MPI_Datatype type;
{
MPI_BXOR ( invec, inoutvec, len, type );
}



#endif
