/*
 *  $Id: scan.c,v 1.12 1994/02/02 21:05:26 lusk Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: scan.c,v 1.12 1994/02/02 21:05:26 lusk Exp $";
#endif /* lint */

#include <mpi.h>
#include "coll.h"

/*@

MPI_Scan - 

@*/
int MPI_Scan ( sendbuf, recvbuf, count, datatype, op, comm )
void             *sendbuf;
void             *recvbuf;
int               count;
MPI_Datatype      datatype;
void            (*op)();
MPI_Comm          comm;
{
  MPI_Status status;
  int        rank, size, extent;
  int        errno = MPI_SUCCESS;


  /* Check for invalid arguments */
  if ( ( (comm            == (MPI_Comm)0)   && (errno = MPI_ERR_COMM_NULL) ) ||
       ( (op              == (void (*)())0) && (errno = MPI_ERR_ROOT) )      || 
       ( (sendbuf         == (void *)0)     && (errno = MPI_ERR_BUFFER) )    ||
       ( (recvbuf         == (void *)0)     && (errno = MPI_ERR_BUFFER) )    ||
       ( (count           <= 0)             && (errno = MPI_ERR_COUNT) )     ||
       ( (comm->comm_type == MPIR_INTER)    && (errno = MPI_ERR_COMM_INTER) ) )
    return (errno); 
  /* We also need to check that the datatype is a basic type? */

  /* Get my rank & size and switch communicators to the hidden collective */
  MPI_Comm_size ( comm, &size );
  MPI_Comm_rank ( comm, &rank );
  MPI_Type_extent ( datatype, &extent );
  comm = comm->comm_coll;

  /* Initialize the op */
  MPIR_Init_coll_op ( sendbuf, count, datatype, op, rank );

  /* If the recv and send buffers are equivalent */
  if (recvbuf == sendbuf) {
    if (rank > 0) {
      void *buffer = (void *)MALLOC(extent * count);
      MPI_Recv(buffer,count,datatype,rank-1,MPIR_SCAN_TAG,comm,&status);
      (*op)(buffer, recvbuf, count, datatype); 
      FREE(buffer);
    }
  }
  else  /* send and recv buffers are different */
    if (rank > 0) {
      MPI_Recv(recvbuf,count,datatype,rank-1,MPIR_SCAN_TAG,comm,&status);
      (*op)(sendbuf, recvbuf, count, datatype); 
    }
    else {
      MPI_Sendrecv(sendbuf,count,datatype,rank, MPIR_SCAN_TAG,
		   recvbuf,count,datatype,rank, MPIR_SCAN_TAG,
		   comm, &status);
    }

  /* send the letter to destination */
  if (rank < (size-1)) 
    MPI_Send(recvbuf,count,datatype,rank+1,MPIR_SCAN_TAG,comm);

  return(errno);
}
