/*
 *  $Id: u_reducea.c,v 1.9 1994/01/27 16:10:10 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: u_reducea.c,v 1.9 1994/01/27 16:10:10 doss Exp $";
#endif /* lint */

#include <mpi.h>
#include "coll.h"

/*@

MPI_User_reducea - 

algorithm:  The way this function is done is horrible.  Everybody
reduces to node 0, the node 0 sends the result to node root.  Yuck.
This is done to meet the associativity requirement.  Change this soon.
@*/

MPI_User_reducea ( sendbuf, recvbuf, count, datatype, op, root, comm )
void             *sendbuf;
void             *recvbuf;
int               count;
MPI_Datatype      datatype;
void            (*op)();
int               root;
MPI_Comm          comm;
{
  MPI_Status status;
  int        size, rank;
  int        errno = MPI_SUCCESS;

  /* Check for invalid arguments */
  if ( ( (comm            == (MPI_Comm)0) && (errno = MPI_ERR_COMM_NULL) ) ||
       ( (root            <  0)           && (errno = MPI_ERR_ROOT) )      || 
       ( (comm->comm_type == MPIR_INTER)  && (errno = MPI_ERR_COMM_INTER) ) )
    return (errno); 
  
  /* Is root within the communicator? */
  MPI_Comm_size ( comm, &size );
  if ( root >= size )
    return (errno);

  /* Reduce to node 0.  This takes care of the associativity part */
  MPI_Comm_rank ( comm, &rank );
  MPI_User_reduce ( sendbuf, recvbuf, count, datatype, op, 0, comm );
  if (rank == 0) 
    MPI_Send ( recvbuf, count, datatype, root, MPIR_USER_REDUCEA_TAG, comm );

  if (rank == root) 
    MPI_Recv ( recvbuf, count, datatype, 0, MPIR_USER_REDUCEA_TAG, comm, &status );
  
  return (errno);
}


