/*
 *  $Id: group_diff.c,v 1.5 1994/02/02 21:12:27 lusk Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

#define MPIR_UNMARKED 0
#define MPIR_MARKED   1

/*@

MPI_Group_difference - 

@*/
int MPI_Group_difference ( group1, group2, group_out )
MPI_Group group1, group2, *group_out;
{
  int        i, j, global_rank;
  MPI_Group  new_group;
  int        n;

  /* Check for NULL groups */
  if ( group1 == (MPI_Group)0 ) {
    (*group_out) = (MPI_Group)0;
    return (MPI_SUCCESS);
  }
  if ( group2 == (MPI_Group)0 ) {
    (void) MPIR_GROUP_dup ( group1, group_out );
    return (MPI_SUCCESS);
  }
  
  /* Create the new group */
  new_group = (*group_out) = NEW(MPIR_GROUP);
  new_group->ref_count     = 1;
  new_group->local_rank    = MPI_UNDEFINED;
  new_group->set_mark      = (int *)0;

  /* Set the number in the difference */
  n = group1->np;

  /* Allocate set marking space for group1 if necessary */
  if (group1->set_mark == NULL)
    group1->set_mark = (int *) MALLOC( group1->np * sizeof(int) );

  /* Mark the difference */
  for ( i=0; i<group1->np; i++ ) {
    group1->set_mark[i] = MPIR_MARKED;
    for ( j=0; j<group2->np; j++ ) 
      if ( group1->lrank_to_grank[i] == group2->lrank_to_grank[j] ) {
        group1->set_mark[i] = MPIR_UNMARKED;
        n--;
        break;
      }
  }

  /* Alloc memory for lrank_to_grank array */
  new_group->np             = n;
  new_group->lrank_to_grank = (int *) MALLOC( n * sizeof(int) );
    
  /* Fill in the space */
  for ( n=0, i=0; i<group1->np; i++ ) 
    if ( (group1->set_mark[i]==MPIR_MARKED) && (n < new_group->np) ) 
      new_group->lrank_to_grank[n++] = group1->lrank_to_grank[i];

  /* Find the local rank */
  MPID_myrank ( &global_rank );
  for( i=0; i<new_group->np; i++ )
    if ( global_rank == new_group->lrank_to_grank[i] ) {
      new_group->local_rank = i;
      break;
    }

  /* Determine the previous and next powers of 2 for this group */
  MPIR_Powers_of_2 ( new_group->np, &(new_group->N2_next), &(new_group->N2_prev) );

  return (MPI_SUCCESS);
}
