/*
 *  $Id: group_tranks.c,v 1.5 1994/02/02 21:25:02 lusk Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

/*@

MPI_Group_translate_ranks - 

@*/
int MPI_Group_translate_ranks ( group_a, n, ranks_a, group_b, ranks_b )
MPI_Group group_a;
int n;
int       *ranks_a;
MPI_Group  group_b;
int       *ranks_b;
{
  int i, j;
  int pid_a, rank_a;

  /* Check for valid number of ranks, valid ranks_a, and valid ranks_b */
  if ( n<=0 )
    return (MPI_ERR_UNKNOWN);

  /* Set ranks_b array to MPI_UNDEFINED */
  for ( i=0; i<n; i++ )
    ranks_b[i] = MPI_UNDEFINED;
 
  /* If group is NULL, exit with MPI_SUCCESS, ranks_b is MPI_UNDEFINED */
  if ( (group_a == (MPI_Group)0) || (group_b == (MPI_Group)0) ||
       (ranks_a == (int *)0) ) 
    return (MPI_ERR_UNKNOWN);

  /* Translate the ranks in ranks_a to ranks_b */
  for ( i=0 ; i<n; i++ ) {
    if ( ((rank_a = ranks_a[i]) >= group_a->np) || (rank_a < 0) ) 
      break;
    pid_a = group_a->lrank_to_grank[rank_a];
    for ( j=0; j<group_b->np; j++ ) 
      if ( pid_a == group_b->lrank_to_grank[j] ) {
        ranks_b[i] = j;
        break;
      }
  }
  return (MPI_SUCCESS);
}
