/*
 *  $Id: cart_sub.c,v 1.5 1994/01/25 21:58:02 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

/*@

MPI_Cart_sub - used to partition a communicator into subgroups which 
               form lower-dimensional cartesian subgrids

@*/
int MPI_Cart_sub ( comm, remain_dims, comm_new )
MPI_Comm  comm;
int      *remain_dims;
MPI_Comm *comm_new;
{
  int i, j, ndims;
  int remain_total = 1;
  int num_remain_dims = 0;
  int rank, color, key;
  int errno = MPI_SUCCESS;

  /* Check for valid arguments */
  if ( ((comm                == (MPI_Comm)0)&&(errno = MPI_ERR_COMM_NULL)) ||
       ((comm->topology.type != MPI_CART)   &&(errno = MPI_ERR_TOPOLOGY))  ||
       ((remain_dims         == (int *)0)   &&(errno = MPI_ERR_NULL))      ||
       ((comm_new          == (MPI_Comm *)0)&&(errno = MPI_ERR_NULL))       )
    return (errno);
  
  /* Determine the remaining dimension total */
  ndims = comm->topology.tag.cart.ndims;
  for ( i=0; i<ndims; i++ )
    if ( remain_dims[i] ) {
      num_remain_dims++;
      remain_total *= comm->topology.tag.cart.dims[i];
    }

  /* Check for special case */
  if ( num_remain_dims == 0 ) {
    (*comm_new) = (MPI_Comm)0;
    return (errno);
  }

  /* Split the old communicator */
  MPI_Comm_rank ( comm, &rank );
  color = rank / remain_total;
  key = rank % remain_total;
  MPI_Comm_split ( comm, color, key, comm_new );

  /* Store topology information in new communicator */
  if ( (*comm_new) != (MPI_Comm)0 ) {
    (*comm_new)->topology.type                    = MPI_CART;
    (*comm_new)->topology.tag.cart.nnodes         = remain_total;
    (*comm_new)->topology.tag.cart.ndims          = num_remain_dims;
    (*comm_new)->topology.tag.cart.dims           = 
      (int *)MPIR_MALLOC( sizeof(int) * 3 * num_remain_dims );
    (*comm_new)->topology.tag.cart.periods        = 
      (*comm_new)->topology.tag.cart.dims + num_remain_dims * 3;
    (*comm_new)->topology.tag.cart.position       = 
      (*comm_new)->topology.tag.cart.periods + num_remain_dims * 3;
    for ( i=j=0; i<ndims; i++ )
      if ( remain_dims[i] ) {
		(*comm_new)->topology.tag.cart.dims[j]    = comm->topology.tag.cart.dims[i];
		(*comm_new)->topology.tag.cart.periods[j] = comm->topology.tag.cart.periods[i];
		j++;
	  }
  
    /* Compute my position */
    MPI_Comm_rank ( (*comm_new), &rank );
    for ( i=0; i < num_remain_dims; i++ ) {
      remain_total = remain_total / (*comm_new)->topology.tag.cart.dims[i];
      (*comm_new)->topology.tag.cart.position[i]  = rank / remain_total;
      rank = rank % remain_total;
    }
  }
  return (errno);
}

