/*
 *  $Id: get_graph.c,v 1.6 1994/02/02 21:29:10 lusk Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

/*@

MPI_Get_graph - retrieve graph topology information associated with a 
                communicator

@*/
int MPI_Get_graph ( comm, maxindex, maxedges, index, edges )
MPI_Comm comm;
int maxindex, maxedges;
int *index, *edges;
{
  int i, num;
  int *array;
  int errno = MPI_SUCCESS;

  /* Check for valid arguments */
  if (((comm                == (MPI_Comm)0) && (errno = MPI_ERR_COMM_NULL)) ||
      ((comm->topology.type != MPI_GRAPH)   && (errno = MPI_ERR_TOPOLOGY))   )
    return (errno);

  /* Get index */
  num = comm->topology.tag.graph.nnodes;
  array = comm->topology.tag.graph.index;
  if ( index != (int *)0 )
    for ( i=0; (i<maxindex) && (i<num); i++ )
      (*index++) = (*array++);

  /* Get edges */
  num = comm->topology.tag.graph.nedges;
  array = comm->topology.tag.graph.edges;
  if ( edges != (int *)0 )
    for ( i=0; (i<maxedges) && (i<num); i++ )
      (*edges++) = (*edges++);

  return (errno);
}
